#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE 600
#endif
#include <pthread.h>
#include <sys/uio.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>

#include "Spooler.h"

using namespace SafeSpool;



// Invariant:
//
//               +---------+
//               |         |
//               |         |
//               |         |
//               |         |
//               |         |
// low_mark    ->|         |
//               |         |
//               |         |
//               |         |
//               |         |
//               |         |
// cycle_size  ->|         |
//               |         |
//               |         |      +---------+  <- overflow_offset
//               |         |      |         |
// high_mark   ->|         |      |         |
//                            
// (files grow downward)


//Actual spooler implementation
struct SafeSpool::Spooler_impl {
	Spooler *spooler;
	char primary_filename[1024];
	char overflow_filename[1024];
	
	size_t cycle_size;
	
	int fd_primary;
	int fd_primary_read;
	int fd_overflow;
	
	size_t primary_low_mark;
	size_t primary_high_mark;
	
	size_t overflow_offset;
	
	bool terminate;
	stop_action_t stop_action;
	pthread_mutex_t mtx;
	pthread_cond_t cond_wakeup;
	pthread_t sender_thread;
	
	bool init(Spooler *spooler_, const char *basename, size_t cycle_size_);
	void finish(SafeSpool::stop_action_t stop_action);
	
	bool queueBatch(const Spooler::Item *item, int items);
	bool writeItems(int fd, const Spooler::Item *item, int item);
	bool storeItems(const Spooler::Item *item, int items);
	bool startOverflow(size_t old_primary_offset);
	bool setLowMark(size_t o);
	bool switchFiles();
	
	void senderThreadFunc();
};


extern "C" {
static void *senderThreadFunc_trampoline(void *pv) {
	Spooler_impl *impl = (Spooler_impl*)pv;
	impl->senderThreadFunc();
	return 0;
}
}

bool Spooler_impl::init(Spooler *spooler_, const char *basename, size_t cycle_size_) {
	spooler = spooler_;
	
	strcpy(primary_filename,basename);
	strcpy(overflow_filename,basename);
	strcat(overflow_filename,".overflow");
	
	cycle_size = cycle_size_;
	
	fd_primary = open(primary_filename, O_WRONLY|O_CREAT|O_APPEND, 0600);
	if(fd_primary<0) goto fail0;
	fd_primary_read = open(primary_filename, O_RDONLY);
	if(fd_primary_read<0) goto fail1;
	fd_overflow = -1;
	
	primary_low_mark = 0;
	primary_high_mark = lseek(fd_primary, 0, SEEK_END);
	overflow_offset = (size_t)-1;
	
	terminate = false;
	if(pthread_mutex_init(&mtx,0)!=0) goto fail2;
	if(pthread_cond_init(&cond_wakeup,0)!=0) goto fail3;
	if(pthread_create(&sender_thread,0,senderThreadFunc_trampoline,this)!=0) goto fail4;
	
	return true;

fail4:
	pthread_cond_destroy(&cond_wakeup);
fail3:
	pthread_mutex_destroy(&mtx);
fail2:
	close(fd_primary_read);
fail1:
	close(fd_primary);
fail0:
	return false;
}


//Stop the sender thread and clean up
void Spooler_impl::finish(stop_action_t stop_action_) {
	pthread_mutex_lock(&mtx);
	stop_action = stop_action_;
	terminate = true;
	pthread_mutex_unlock(&mtx);
	pthread_cond_signal(&cond_wakeup);
	pthread_join(sender_thread,0);
	pthread_mutex_destroy(&mtx);
	pthread_cond_destroy(&cond_wakeup);
	
	close(fd_primary);
	close(fd_primary_read);
	if(fd_overflow!=-1) {
		close(fd_overflow);
		remove(overflow_filename);
	}
}


//Put items into the queue
bool Spooler_impl::queueBatch(const Spooler::Item *item, int items) {
	pthread_mutex_lock(&mtx);
	bool rc = storeItems(item,items);
	pthread_mutex_unlock(&mtx);
	if(!rc)
		return false;
	pthread_cond_signal(&cond_wakeup);
	return true;
}

//Write items into a file
bool Spooler_impl::writeItems(int fd, const Spooler::Item *item, int items) {
	size_t total_size=0;
	for(int k=0; k<items; k++)
		total_size += sizeof(item[k].sz) + item[k].sz;
	if(total_size<=8192) {
		//Do gather ourselves
		char buf[8192];
		char *p=buf;
		for(int i=0; i<items; i++) {
			memcpy(p, &item[i].sz, 4);
			p+=4;
			memcpy(p, item[i].buf, item[i].sz);
			p += item[i].sz;
		}
		if(write(fd,buf,total_size)!=(long)total_size)
			return false;
		
		return true;
	} else {
		struct iovec iov[IOV_MAX];
		const int iovs_per_item=2;
		const int items_per_send = IOV_MAX/iovs_per_item;
		for(int i=0; i<items; i+= items_per_send) {
			int iov_i=0;
			for(int j=i; j<items && j<i+items_per_send; j++) {
				iov[iov_i].iov_base = (void*)&item[j].sz;
				iov[iov_i].iov_len = sizeof(item[j].sz);
				iov[iov_i+1].iov_base = (void*)item[j].buf;
				iov[iov_i+1].iov_len = item[j].sz;
				iov_i+=2;
			}
			if(writev(fd, iov, iov_i)==-1)
				return false;
		}
		return true;
	}

}


//Store items into primary file, start overflow if necessary and store them there too
bool Spooler_impl::storeItems(const Spooler::Item *item, int items) {
	size_t old_primary_offset = lseek(fd_primary,0,SEEK_END); //CUR ?
	
	if(!writeItems(fd_primary, item, items)) return false;
	
	primary_high_mark = lseek(fd_primary,0,SEEK_CUR);
	if(primary_high_mark > cycle_size) {
		if(fd_overflow==-1)
			if(!startOverflow(old_primary_offset)) return false;
		if(!writeItems(fd_overflow, item, items)) return false;
	}
	return true;
}


//Create an overflow file
bool Spooler_impl::startOverflow(size_t old_primary_offset) {
	fd_overflow = open(overflow_filename, O_WRONLY|O_CREAT|O_TRUNC|O_APPEND, 0600);
	if(fd_overflow<0) return false;
	overflow_offset = old_primary_offset;
	return true;
}


//Set the low-mark. Truncate or switch files if possible
bool Spooler_impl::setLowMark(size_t o) {
//printf("p-low:%u p-high:%u o-o:%u  \r",primary_low_mark,primary_high_mark,overflow_offset);
	primary_low_mark = o;
	if(primary_low_mark==primary_high_mark) {
		//empty
//printf("\ntruncate\n");
		ftruncate(fd_primary,0);
		lseek(fd_primary,0,SEEK_SET);
		lseek(fd_primary_read,0,SEEK_SET);
		if(fd_overflow!=-1) {
			ftruncate(fd_overflow,0);
			lseek(fd_overflow,0,SEEK_SET);
		}
		primary_low_mark = primary_high_mark = 0;
		overflow_offset = (size_t)-1;
	} else {
		if(primary_low_mark >= overflow_offset)
			switchFiles();
	}
	return true;
}


bool Spooler_impl::switchFiles() {
//printf("\nswitch\n");
	rename(overflow_filename, primary_filename);
	close(fd_primary);
	close(fd_primary_read);
	
	fd_primary = fd_overflow;
	fd_overflow = -1;
	primary_low_mark = primary_low_mark - overflow_offset;
	primary_high_mark = primary_high_mark - overflow_offset;
	overflow_offset = (size_t)-1;
	
	fd_primary_read = open(primary_filename,O_RDONLY);
	lseek(fd_primary_read, primary_low_mark, SEEK_SET);
	
	return true;
}


static void analyzeBytes(const char *buf, long bytes,
                         Spooler::Item *item, int *items, long *usable_bytes)
{
	*items = 0;
	long o=0;
	while(o<bytes-4) {
		size_t sz;
		memcpy(&sz, buf+o, sizeof(size_t));
		size_t total = sizeof(sz) + sz;
		if(o+total > (size_t)bytes)
			break;
		item->buf = buf+o;
		item->sz = sz;
		item++;
		++*items;
		o += total;
	}
	
	*usable_bytes = o;
}

void Spooler_impl::senderThreadFunc() {
	const int bufsize=65536;
	const int max_items = bufsize/sizeof(size_t);
	Spooler::Item *item = new Spooler::Item[max_items];
	char *buf = new char[bufsize];
	bool first_loop=true;
	while(!terminate) {
		pthread_mutex_lock(&mtx);
		if(terminate && stop_action==stop_now) {
			pthread_mutex_unlock(&mtx);
			break;
		}
		if(!first_loop)
			pthread_cond_wait(&cond_wakeup,&mtx);
		pthread_mutex_unlock(&mtx);
		first_loop = false;
		if(terminate) break;
		
		size_t low = lseek(fd_primary_read,0,SEEK_CUR);
		long bytes_read = read(fd_primary_read, buf, bufsize);
		if(bytes_read<0)
			break; //error
		if(bytes_read==0)
			continue; //spurious wakeup
		long usable_bytes;
		int items;
		analyzeBytes(buf,bytes_read,item,&items,&usable_bytes);
		if(usable_bytes==0)
			continue; //a bit strange
		size_t high = low + usable_bytes;
		if(usable_bytes!=bytes_read)
			lseek(fd_primary_read, low+usable_bytes, SEEK_SET);
		
		if(!spooler->forwardBatch(item,items))
			break;
		
		pthread_mutex_lock(&mtx);
		setLowMark(high);
		pthread_mutex_unlock(&mtx);
	}
	delete[] buf;
	delete[] item;
}




//-----------------------------------------------------------------------------


Spooler::Spooler(const char *basename, size_t cycle_size) {
	impl = new Spooler_impl;
	if(!impl->init(this,basename,cycle_size)) {
		delete impl;
		impl = 0;
		//throw something
	}
}


Spooler::~Spooler() {
	stop();
}


void Spooler::stop(stop_action_t stop_action) {
	if(impl) {
		impl->finish(stop_action);
		delete impl;
		impl = 0;
	}
}


bool Spooler::queueBatch(const Item *item, int items) {
	if(impl)
		return impl->queueBatch(item,items);
	else
		return false;
}

//This is marked a pure virtual but unde rdestruction when the user forgot to
//call stop() it may end getting called
bool Spooler::forwardBatch(const Item */*item*/, int /*items*/) {
	return false;
}
